.. Known projects

.. PROJECTNAME placeholders
.. _PROJECTNAME: http://nipy.org
.. _`PROJECTNAME github`: https://github.com/nipy
.. _`PROJECTNAME mailing list`: https://mail.python.org/mailman/listinfo/neuroimaging

.. numpy
.. _numpy: http://www.numpy.org
.. _`numpy github`: https://github.com/numpy/numpy
.. _`numpy mailing list`: http://mail.scipy.org/mailman/listinfo/numpy-discussion

.. scipy
.. _scipy: https://www.scipy.org
.. _`scipy github`: https://github.com/scipy/scipy
.. _`scipy mailing list`: http://mail.scipy.org/mailman/listinfo/scipy-dev

.. nipy
.. _nipy: http://nipy.org/nipy
.. _`nipy github`: https://github.com/nipy/nipy
.. _`nipy mailing list`: https://mail.python.org/mailman/listinfo/neuroimaging

.. ipython
.. _ipython: https://ipython.org
.. _`ipython github`: https://github.com/ipython/ipython
.. _`ipython mailing list`: http://mail.scipy.org/mailman/listinfo/IPython-dev

.. dipy
.. _dipy: http://nipy.org/dipy
.. _`dipy github`: https://github.com/Garyfallidis/dipy
.. _`dipy mailing list`: https://mail.python.org/mailman/listinfo/neuroimaging

.. nibabel
.. _nibabel: http://nipy.org/nibabel
.. _`nibabel github`: https://github.com/nipy/nibabel
.. _`nibabel mailing list`: https://mail.python.org/mailman/listinfo/neuroimaging

.. marsbar
.. _marsbar: http://marsbar.sourceforge.net
.. _`marsbar github`: https://github.com/matthew-brett/marsbar
.. _`MarsBaR mailing list`: https://lists.sourceforge.net/lists/listinfo/marsbar-users
